#include "fastlz.h"
#include <cstddef>
#include <cstdlib>
#include <malloc.h>
#include <nds.h>
#include <filesystem.h>
#include <nds/arm9/decompress.h>
#include <nds/bios.h>
#include <nds/dma.h>
#include <nds/system.h>
#include <nds/timers.h>
#include <stdio.h>

#define TIMER_SPEED (BUS_CLOCK/1024)

void wait_forever(const char* msg);
size_t LoadFile(const char* file, unsigned char **buffer);
void readNextBatch(FILE *input, void *output);
void vbl_handler();

class CircularQueue {
public:
  int front, rear, size, current;
  CircularQueue(int s) {
    size = s;
    front = rear = -1;
		current = 0;
  }

  bool isFull() {
    return (front == 0 && rear == size - 1) || (rear == (front - 1) % (size - 1));
  }

  bool isEmpty() {
    return front == -1;
  }

  void push() {
    if (isFull()) {
      fprintf(stderr, "Queue is full. Cannot push\n");
      return;
    }

    if (isEmpty()) {
      front = rear = 0;
    } else {
      rear = (rear + 1) % size;
    }

		current++;
  }

  void pop() {
    if (isEmpty()) {
      fprintf(stderr, "Queue is empty. Cannot pop\n");
      return;
    }

    if (front == rear) {
      front = rear = -1;
    } else {
      front = (front + 1) % size;
    }

		current--;
  }
};

static uint8_t chunk[256*192*4*7];
static int chunk_counter = 0;
static CircularQueue queue(7);
static FILE *in = NULL;
static int bgp;

int main(void) {

	videoSetMode(MODE_5_3D);
	videoSetModeSub(MODE_0_2D);

	vramSetBankA(VRAM_A_MAIN_BG);
	vramSetBankB(VRAM_B_TEXTURE);
	vramSetBankF(VRAM_F_TEX_PALETTE);

	consoleDemoInit();
	consoleDebugInit(DebugDevice_NOCASH);

	nitroFSInit(NULL);

	in = fopen("nitro:/image.bin", "rb");
	if (in == NULL) wait_forever("cannot load nitro:/image.bin");

	// unsigned char* bmp = NULL;
	// unsigned char* bmp_decompress = (unsigned char*)malloc(256*192);
	// size_t bmp_len = LoadFile("nitro:/image.bin", &bmp);

	// // size_t bmp_decompress_len = bmp[0] | bmp[1] << 8;
	// size_t bmp_decompress_len = 0;

	// memcpy(&bmp_decompress_len, bmp, sizeof(uint32));
	// fprintf(stderr, "%i %i\n", bmp_decompress_len, bmp_len);
	// fastlz_decompress(&bmp[4], bmp_decompress_len, bmp_decompress, 256*192);

	unsigned char* pal = NULL;
	size_t pal_len = LoadFile("nitro:/palette.bin", &pal);

	bgp = bgInit(3, BgType_Bmp8, BgSize_B8_256x256, 0, 0);

	dmaCopy(pal, BG_PALETTE, pal_len);
	// // DC_FlushRange(bg, bg_len);
	// dmaCopy(bmp_decompress, bgGetGfxPtr(bgp), 256*192);
	// nocashMessage("done\n");


	// fprintf(stderr, "start debug\n");
	// int cframe = 0;
	// int hframe = 0;
	// int ticks = 0;


	// while(!feof(in)) 
	// {
	// 	timerStart(0, ClockDivider_1024, 0, NULL);
	// 	readNextBatch(in , chunk);
	// 	ticks = timerStop(0);
	// 	int milis = ((ticks%TIMER_SPEED)*1000) /TIMER_SPEED;
		
	// 	if (milis > 60) fprintf(stderr, "%i\n", cframe);
	// 	cframe++;
	// }

	// fprintf(stderr, "%i\n", hframe);


	fseek(in, 0, SEEK_SET);

	fprintf(stderr, "read\n");
	while(!queue.isFull()) {
		queue.push();
		readNextBatch(in, &chunk[256*192*4 * queue.rear]);
		fprintf(stderr, "read %i\n", queue.rear);
		// if (queue.rear == 1) break;
	}

	while(1) {

		irqSet(IRQ_VBLANK, vbl_handler);

		if (!queue.isFull()) {
			queue.push();
			readNextBatch(in, &chunk[256*192*4 * queue.rear]);
			// fprintf(stderr, "read %i\n", queue.rear);
		}
	}

	return 0;
}

void vbl_handler()
{
	// if (queue.isEmpty() && chunk_counter == 0) {
	// 	return;
	// 	fprintf(stderr, "NO QUEUE!!!!!");
	// }

	dmaCopy(&chunk[256*192*4 * queue.front + 256*192*chunk_counter], bgGetGfxPtr(bgp), 256*192);
	fprintf(stderr, "count: %i, %i\n ", chunk_counter, queue.current);
	chunk_counter++;

	if (chunk_counter >= 4) {
		if (!queue.isEmpty()) {
			queue.pop();
			chunk_counter = 0;
		} else {
			chunk_counter = 4;
			fprintf(stderr, "NO QUEUE!!!!!");
		}
	}
}

void readNextBatch(FILE *input, void *output)
{
	uint32_t chunk_len;
	fread(&chunk_len, sizeof(uint32_t), 1, input);

	bool isRLE = chunk_len & 0x01;
	chunk_len = chunk_len >> 1;

	uint8_t buffer[chunk_len];
	// uint8_t *buffer2 = (uint8_t*)malloc(256*192*4);
	fread(buffer, 1, chunk_len, input);

	// DC_FlushRange(output, 256*192*4);
	if (isRLE) 
	{
		// fprintf(stderr, "rle\n");
		decompress(buffer, output, RLE);
		// dmaCopy(buffer2, output, 256*192*4);
	}
	else 
	{
		// fprintf(stderr, "fastlz\n");
		fastlz_decompress(buffer, chunk_len, output, 256*192*4);
	}
}

size_t LoadFile(const char* file, unsigned char **buffer)
{
	FILE *in = fopen(file, "rb");
	if (in == NULL) wait_forever("cannot load file!");

	fseek(in, 0, SEEK_END);
	long file_size = ftell(in);
	rewind(in);

	*buffer = (unsigned char*)malloc(file_size);
	if (*buffer == NULL) wait_forever("Failed to malloc!");

	size_t bytes_read = fread(*buffer, 1, file_size, in);
	if (bytes_read != file_size)
	{
		fclose(in);
		free(*buffer);
		wait_forever("Failed to read");
	}

	return bytes_read;
}

void wait_forever(const char* msg)
{
	fprintf(stdout, "%s", msg);
	fprintf(stderr, "%s", msg);
  while (1) swiWaitForVBlank();
}